from __future__ import annotations

from .coast import VtkCoastLineSource
from .mesh import VtkMeshReader
from .segment import VtkSegmentReader

__all__ = [
    "VtkCoastLineSource",
    "VtkMeshReader",
    "VtkSegmentReader",
]



from __future__ import annotations

from pathlib import Path

import h5py
import numpy as np
from pyproj import Geod
from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase
from vtkmodules.vtkCommonCore import vtkFloatArray, vtkPoints
from vtkmodules.vtkCommonDataModel import vtkCellArray, vtkPolyData

from parsli.utils import earth

APPROXIMATE_SCALING = 111.0


def km_to_degrees_lon(km, lat):
    return km / (APPROXIMATE_SCALING * np.cos(np.deg2rad(lat)))


def km_to_degrees_lat(km):
    return km / APPROXIMATE_SCALING


def get_segment_bottom_lon_lat(lon1, lat1, lon2, lat2, locking_depth, dip):
    # Get segment azimuths
    azimuth = Geod(ellps="WGS84").inv(lon1, lat1, lon2, lat2)[0]

    # Get segment length
    length_km = locking_depth / np.tan(np.deg2rad(dip))

    # Get longitude and latitude spans
    delta_lon_km = length_km * np.cos(np.deg2rad(azimuth))
    delta_lat_km = -length_km * np.sin(np.deg2rad(azimuth))

    # Get average latitude
    avg_lat = (lat1 + lat2) / 2.0
    delta_lon = km_to_degrees_lon(delta_lon_km, avg_lat)
    delta_lat = km_to_degrees_lat(delta_lat_km)

    # Calculate approximate longitude and latitudes of lower vertices
    lon3 = lon1 + delta_lon
    lon4 = lon2 + delta_lon
    lat3 = lat1 + delta_lat
    lat4 = lat2 + delta_lat
    return lon3, lat3, lon4, lat4


class EarthLocation:
    __slots__ = ("lat", "lon")

    def __ilshift__(self, other):
        self.lat = other.lat
        self.lon = other.lon

    def flip(self):
        self.lon *= -1
        self.lat *= -1

    def interpolate_from(self, start_lon, start_lat, end_lon, end_lat, distance):
        self.lon, self.lat = earth.interpolate(
            start_lon, start_lat, end_lon, end_lat, distance
        )

    def __repr__(self):
        return f"Longitude: {self.lon}, Latitude: {self.lat}"


FIELD_COLS = {
    "strike_slip": 68,
    "dip_slip": 69,
    "tensile_slip": 70,
}
FIELD_NAMES = list(FIELD_COLS.keys())


class QuadCell:
    __slots__ = (
        "dip",
        "end",
        "latitude_bnds",
        "locking_depth",
        "longitude_bnds",
        "normal",
        "point_a",
        "point_b",
        "start",
    )

    def __init__(self, longitude_bnds, latitude_bnds):
        self.longitude_bnds = longitude_bnds
        self.latitude_bnds = latitude_bnds
        self.start = EarthLocation()
        self.point_a = EarthLocation()
        self.point_b = EarthLocation()
        self.end = EarthLocation()
        self.normal = EarthLocation()

    def update(self, row):
        if row[34]:
            # skip cell if column 34 is true
            return False

        if row[0] >= row[2]:
            self.start.lon = row[0]
            self.start.lat = row[1]
            self.end.lon = row[2]
            self.end.lat = row[3]
        else:
            self.end.lon = row[0]
            self.end.lat = row[1]
            self.start.lon = row[2]
            self.start.lat = row[3]

        if (
            self.start.lon < self.longitude_bnds[0]
            or self.end.lon > self.longitude_bnds[1]
            or self.start.lat < self.latitude_bnds[0]
            or self.end.lat > self.latitude_bnds[1]
        ):
            # print(f"skip {self.start} {self.end}")
            return False

        self.dip = row[4]
        self.locking_depth = row[14]

        lon3, lat3, lon4, lat4 = get_segment_bottom_lon_lat(
            self.start.lon,
            self.start.lat,
            self.end.lon,
            self.end.lat,
            self.locking_depth,
            self.dip,
        )

        self.point_a.lon = lon3
        self.point_a.lat = lat3
        self.point_b.lon = lon4
        self.point_b.lat = lat4

        return [(k, row[FIELD_COLS[k]]) for k in FIELD_NAMES]


class VtkSegmentReader(VTKPythonAlgorithmBase):
    def __init__(self):
        VTKPythonAlgorithmBase.__init__(
            self,
            nInputPorts=0,
            nOutputPorts=1,
            outputType="vtkPolyData",
        )
        self._file_name = None
        self._proj_spherical = True
        self._longitude_bnds = [0, 360]
        self._latitude_bnds = [-90, 90]

    @property
    def field_names(self):
        return FIELD_NAMES

    @property
    def file_name(self):
        return self._file_name

    @file_name.setter
    def file_name(self, path):
        self._file_name = Path(path)
        if not self._file_name.exists():
            msg = f"Invalid file path: {self._file_name.resolve()}"
            raise ValueError(msg)

        self.Modified()

    @property
    def spherical(self):
        return self._proj_spherical

    @spherical.setter
    def spherical(self, value):
        if self._proj_spherical != value:
            self._proj_spherical = value
            self.Modified()

    @property
    def longitude_bnds(self):
        return self._longitude_bnds

    @longitude_bnds.setter
    def longitude_bnds(self, value):
        if self._longitude_bnds != value:
            self._longitude_bnds = value
            self.Modified()

    @property
    def latitude_bnds(self):
        return self._latitude_bnds

    @latitude_bnds.setter
    def latitude_bnds(self, value):
        if self._latitude_bnds != value:
            self._latitude_bnds = value
            self.Modified()

    def RequestData(self, _request, _inInfo, outInfo):
        if self._file_name is None or not self._file_name.exists():
            return 1

        # Read file and generate mesh
        output = self.GetOutputData(outInfo, 0)
        vtk_points = vtkPoints()
        vtk_points.SetDataTypeToDouble()
        vtk_polys = vtkCellArray()
        vtk_mesh = vtkPolyData()
        vtk_mesh.points = vtk_points
        vtk_mesh.polys = vtk_polys

        # Projection selection
        insert_pt = earth.insert_spherical if self.spherical else earth.insert_euclidian

        with h5py.File(self._file_name, "r") as hdf:
            cell = QuadCell(self.longitude_bnds, self.latitude_bnds)
            h5_ds = hdf["segment"]
            data_size = h5_ds.shape

            # making a line for now (should move to 4 once quad)
            vtk_points.Allocate(data_size[0] * 2)
            vtk_polys.Allocate(data_size[0] * 5)

            # Create fields and attach to mesh
            vtk_field_arrays = {}
            for name in FIELD_NAMES:
                array = vtkFloatArray()
                array.SetName(name)
                array.Allocate(data_size[0])
                vtk_mesh.cell_data.AddArray(array)
                vtk_field_arrays[name] = array

            for row in h5_ds:
                if fields := cell.update(row):
                    vtk_polys.InsertNextCell(4)
                    vtk_polys.InsertCellPoint(
                        insert_pt(vtk_points, cell.start.lon, cell.start.lat, 0)
                    )
                    vtk_polys.InsertCellPoint(
                        insert_pt(
                            vtk_points,
                            cell.point_a.lon,
                            cell.point_a.lat,
                            cell.locking_depth,
                        )
                    )
                    vtk_polys.InsertCellPoint(
                        insert_pt(
                            vtk_points,
                            cell.point_b.lon,
                            cell.point_b.lat,
                            cell.locking_depth,
                        )
                    )
                    vtk_polys.InsertCellPoint(
                        insert_pt(vtk_points, cell.end.lon, cell.end.lat, 0)
                    )

                    # Add fields values
                    for k, v in fields:
                        vtk_field_arrays[k].InsertNextTuple1(v)

        output.ShallowCopy(vtk_mesh)
        return 1



from __future__ import annotations

from pathlib import Path

from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase
from vtkmodules.vtkCommonDataModel import (
    vtkPartitionedDataSet,
)
from vtkmodules.vtkIOXML import vtkXMLPolyDataReader

BASE_DIRECTORY = Path(__file__).parent.with_name("assets").resolve()
RESIONS = [f.name.split("-")[0] for f in BASE_DIRECTORY.glob("*-spherical.vtp")]


def region_to_full_name(name, use_spherical):
    ext = "-cil-spherical.vtp" if use_spherical else "-cil-euclidean.vtp"
    output = BASE_DIRECTORY / f"{name}{ext}"
    # print(f"{name} => {output}")
    return str(output.resolve())


class VtkCoastLineSource(VTKPythonAlgorithmBase):
    def __init__(self):
        VTKPythonAlgorithmBase.__init__(
            self,
            nInputPorts=0,
            nOutputPorts=1,
            outputType="vtkPartitionedDataSet",
        )
        self._proj_spherical = True
        self._active_regions = set()

    @property
    def spherical(self):
        return self._proj_spherical

    @spherical.setter
    def spherical(self, value):
        if self._proj_spherical != value:
            self._proj_spherical = value
            self.Modified()

    @property
    def available_regions(self):
        return RESIONS

    @property
    def active_regions(self):
        return self._active_regions

    @active_regions.setter
    def active_regions(self, value):
        new_region_set = set(value) & set(RESIONS)
        if self._active_regions != new_region_set:
            self._active_regions = new_region_set
            self.Modified()

    def RequestData(self, _request, _inInfo, outInfo):
        # Read file and generate mesh
        output = self.GetOutputData(outInfo, 0)
        all_meshes = vtkPartitionedDataSet()
        all_meshes.SetNumberOfPartitions(len(self._active_regions))

        # reader
        reader = vtkXMLPolyDataReader()

        # Projection selection
        for idx, name in enumerate(self._active_regions):
            reader.file_name = region_to_full_name(name, self.spherical)
            all_meshes.SetPartition(idx, reader())

        output.ShallowCopy(all_meshes)
        return 1



from __future__ import annotations

import base64
import json
from pathlib import Path

import numpy as np
import vtkmodules.vtkRenderingOpenGL2  # noqa: F401
from vtkmodules.vtkCommonCore import vtkLookupTable, vtkUnsignedCharArray
from vtkmodules.vtkCommonDataModel import (
    vtkDataObject,
    vtkDataSetAttributes,
    vtkImageData,
)
from vtkmodules.vtkFiltersCore import (
    vtkAssignAttribute,
    vtkCellDataToPointData,
    vtkThreshold,
)
from vtkmodules.vtkFiltersGeometry import vtkDataSetSurfaceFilter
from vtkmodules.vtkFiltersModeling import (
    vtkBandedPolyDataContourFilter,
    vtkLoopSubdivisionFilter,
)
from vtkmodules.vtkFiltersVerdict import vtkMeshQuality

# VTK factory initialization
from vtkmodules.vtkInteractionStyle import (
    vtkInteractorStyleSwitch,  # noqa: F401
    vtkInteractorStyleTerrain,
)
from vtkmodules.vtkInteractionWidgets import vtkOrientationMarkerWidget
from vtkmodules.vtkIOImage import vtkPNGWriter
from vtkmodules.vtkRenderingAnnotation import vtkAxesActor
from vtkmodules.vtkRenderingCore import (
    vtkActor,
    vtkColorTransferFunction,
    vtkCompositePolyDataMapper,
    vtkPolyDataMapper,
    vtkRenderer,
    vtkRenderWindow,
    vtkRenderWindowInteractor,
)

# Disable warning
vtkLoopSubdivisionFilter.GlobalWarningDisplayOff()


PRESETS = {
    item.get("Name"): item
    for item in json.loads(Path(__file__).with_name("presets.json").read_text())
}

LUTS = {}


def get_preset(preset_name: str) -> vtkColorTransferFunction:
    if preset_name in LUTS:
        return LUTS[preset_name]

    lut = LUTS.setdefault(preset_name, vtkColorTransferFunction())
    preset = PRESETS[preset_name]
    srgb = np.array(preset["RGBPoints"])
    color_space = preset["ColorSpace"]

    if color_space == "Diverging":
        lut.SetColorSpaceToDiverging()
    elif color_space == "HSV":
        lut.SetColorSpaceToHSV()
    elif color_space == "Lab":
        lut.SetColorSpaceToLab()
    elif color_space == "RGB":
        lut.SetColorSpaceToRGB()
    elif color_space == "CIELAB":
        lut.SetColorSpaceToLabCIEDE2000()

    if "NanColor" in preset:
        lut.SetNanColor(preset["NanColor"])

    # Always RGB points
    lut.RemoveAllPoints()
    for arr in np.split(srgb, len(srgb) / 4):
        lut.AddRGBPoint(arr[0], arr[1], arr[2], arr[3])

    return lut


def set_preset(lut: vtkLookupTable, preset_name: str, n_colors=255):
    colors = get_preset(preset_name)
    min, max = colors.GetRange()
    delta = max - min
    lut.SetNumberOfTableValues(n_colors)
    for i in range(n_colors):
        x = min + (delta * i / n_colors)
        rgb = colors.GetColor(x)
        lut.SetTableValue(i, *rgb)
    lut.Build()


def to_image(lut, samples=255):
    colorArray = vtkUnsignedCharArray()
    colorArray.SetNumberOfComponents(3)
    colorArray.SetNumberOfTuples(samples)

    dataRange = lut.GetRange()
    delta = (dataRange[1] - dataRange[0]) / float(samples)

    # Add the color array to an image data
    imgData = vtkImageData()
    imgData.SetDimensions(samples, 1, 1)
    imgData.GetPointData().SetScalars(colorArray)

    # Loop over all presets
    rgb = [0, 0, 0]
    for i in range(samples):
        lut.GetColor(dataRange[0] + float(i) * delta, rgb)
        r = int(round(rgb[0] * 255))
        g = int(round(rgb[1] * 255))
        b = int(round(rgb[2] * 255))
        colorArray.SetTuple3(i, r, g, b)

    writer = vtkPNGWriter()
    writer.WriteToMemoryOn()
    writer.SetInputData(imgData)
    writer.SetCompressionLevel(6)
    writer.Write()

    writer.GetResult()

    base64_img = base64.standard_b64encode(writer.GetResult()).decode("utf-8")
    return f"data:image/png;base64,{base64_img}"


class SceneManager:
    def __init__(self, server):
        self.server = server

        self._lut = vtkLookupTable()
        set_preset(self._lut, "Fast")

        self.geometries = {}

        self.renderer = vtkRenderer(background=(1.0, 1.0, 1.0))
        self.interactor = vtkRenderWindowInteractor()
        self.render_window = vtkRenderWindow(off_screen_rendering=1)

        self.style = vtkInteractorStyleTerrain()

        self.render_window.AddRenderer(self.renderer)
        self.interactor.SetRenderWindow(self.render_window)
        # self.interactor.GetInteractorStyle().SetCurrentStyleToTrackballCamera()
        self.interactor.SetInteractorStyle(self.style)

        camera = self.renderer.active_camera
        camera.position = (1, 0, 0)
        camera.focal_point = (0, 0, 0)
        camera.view_up = (0, 0, 1)

        self.interactor.Initialize()

        axes_actor = vtkAxesActor()
        self.widget = vtkOrientationMarkerWidget()
        self.widget.SetOrientationMarker(axes_actor)
        self.widget.SetInteractor(self.interactor)
        self.widget.SetViewport(0.85, 0, 1, 0.15)
        self.widget.EnabledOn()
        self.widget.InteractiveOff()

    @property
    def ctrl(self):
        return self.server.controller

    def __getitem__(self, key):
        return self.geometries.get(key)

    def reset_camera_to(self, bounds):
        self.renderer.ResetCamera(bounds)

    @property
    def camera(self):
        return self.renderer.active_camera

    def update_view_up(self, view_up):
        self.renderer.active_camera.view_up = view_up

    def apply_zoom(self, scale):
        self.renderer.active_camera.Zoom(scale)

    @property
    def lut(self):
        return self._lut

    def add_geometry(self, name, source, composite=False):
        item = {"name": name, "source": source, "composite": composite}

        if not composite:
            geometry = vtkDataSetSurfaceFilter(input_connection=source.output_port)
            mapper = vtkPolyDataMapper(
                input_connection=geometry.output_port,
                lookup_table=self.lut,
            )
            mapper.InterpolateScalarsBeforeMappingOn()
            item["geometry"] = geometry
            item["mapper"] = mapper
        else:
            mapper = vtkCompositePolyDataMapper(
                input_connection=source.output_port,
                lookup_table=self.lut,
            )
            mapper.InterpolateScalarsBeforeMappingOn()
            item["mapper"] = mapper

        actor = vtkActor(mapper=mapper)
        item["actor"] = actor
        item["actors"] = [actor]

        self.geometries[name] = item

        self.renderer.AddActor(actor)
        self.renderer.ResetCamera()
        self.render_window.Render()

        self.ctrl.view_update()

        return item

    def add_geometry_with_contour(self, name, source, composite=False):
        # pipeline filters
        quality = vtkMeshQuality()
        quality.SetTriangleQualityMeasureToEdgeRatio()
        threshold = vtkThreshold(
            threshold_function=vtkThreshold.THRESHOLD_LOWER,
            lower_threshold=3.99,  # magic number for vtkLoopSubdivisionFilter
        )
        threshold.SetInputArrayToProcess(
            0, 0, 0, vtkDataObject.FIELD_ASSOCIATION_CELLS, "Quality"
        )
        geometry = vtkDataSetSurfaceFilter()
        cell2point = vtkCellDataToPointData()
        refine = vtkLoopSubdivisionFilter(
            number_of_subdivisions=1
        )  # Adjust subdivision quality
        assign = vtkAssignAttribute()
        assign.Assign(
            "dip_slip",  # Will be overridden later
            vtkDataSetAttributes.SCALARS,
            vtkDataObject.FIELD_ASSOCIATION_POINTS,
        )
        bands = vtkBandedPolyDataContourFilter(generate_contour_edges=1)
        # bands.SetScalarModeToIndex()

        # connect pipeline
        (
            source
            >> quality
            >> threshold
            >> geometry
            >> cell2point
            >> refine
            >> assign
            >> bands
        )
        for_surface = refine

        # bands.Update()
        # print(bands.GetOutputDataObject(0))

        item = {
            "name": name,
            "source": source,
            "composite": composite,
            "cell2point": cell2point,
            "refine": refine,
            "assign": assign,
            "bands": bands,
        }

        if not composite:
            # surface
            mapper = vtkPolyDataMapper(
                input_connection=for_surface.output_port,
                lookup_table=self.lut,
            )
            mapper.InterpolateScalarsBeforeMappingOn()
            item["mapper"] = mapper
            # lines
            mapper_lines = vtkPolyDataMapper(
                input_connection=bands.GetOutputPort(1),
            )
            mapper_lines.SetResolveCoincidentTopologyToPolygonOffset()
            item["mapper_lines"] = mapper_lines
        else:
            # surface
            mapper = vtkCompositePolyDataMapper(
                input_connection=for_surface.output_port,
                lookup_table=self.lut,
            )
            mapper.InterpolateScalarsBeforeMappingOn()
            item["mapper"] = mapper
            # lines
            mapper_lines = vtkCompositePolyDataMapper(
                input_connection=bands.GetOutputPort(1),
                scalar_visibility=0,
            )
            mapper_lines.SetResolveCoincidentTopologyToPolygonOffset()
            item["mapper_lines"] = mapper_lines

        # Surface actor
        actor = vtkActor(mapper=mapper)
        item["actor"] = actor

        # Lines actor
        actor_lines = vtkActor(mapper=mapper_lines)
        actor_lines.property.color = (0, 0, 0)
        actor_lines.property.line_width = 2
        # actor_lines.property.render_line_as_tube = 1
        item["actor_lines"] = actor_lines

        item["actors"] = [actor, actor_lines]

        self.geometries[name] = item

        self.renderer.AddActor(actor)
        self.renderer.AddActor(actor_lines)
        self.renderer.ResetCamera()
        self.render_window.Render()

        self.ctrl.view_update()

        return item



"""
Copyright (c) 2025 Brendan Meade. All rights reserved.

parsli: Earth data viewers
"""

from __future__ import annotations

import logging

__version__ = "0.10.1"
__all__ = ["__version__"]

# Set global config
logging.basicConfig(level=logging.CRITICAL)



from __future__ import annotations

from pathlib import Path

import h5py

BASE_DIRECTORY = Path(__file__).parent.parent.parent.parent
INPUT = "data/model_0000000881_multi.hdf5"


def explore_ds():
    data_file = BASE_DIRECTORY / INPUT

    with h5py.File(data_file, "r") as hdf:

        def print_info(name):
            print(name, "=", hdf[name])

        hdf.visit(print_info)


if __name__ == "__main__":
    explore_ds()



# Let start by reading the line segments
from __future__ import annotations

from parsli.io.sample import SAMPLE_DATASET
from parsli.io.segment import VtkSegmentReader

reader = VtkSegmentReader()
reader.file_name = SAMPLE_DATASET

output = reader()
print(output)
